import json
import sys
from typing import Any, List, Optional

from google.protobuf import json_format
import redis
import schedule

from jetpack import utils
from jetpack._job.job import Job
from jetpack.cmd import is_using_new_cli
from jetpack.config import symbols
from jetpack.config.symbols import Symbol
from jetpack.models.runtime import cronjob_pb2, describe_pb2, job_pb2

# TODO: Change types of both `app` and return type. Return type is DescribeOutput
# but because that protocol buffer is still outside of /proto we are not generating
# type information for it.


# TODO(Landau): Use a framework? Like https://click.palletsprojects.com/en/7.x/
def handle(app: Any = None) -> None:
    run(app=app, cli_args=sys.argv)


# TODO(Landau): Use a framework? Like https://click.palletsprojects.com/en/7.x/
def run(app: Any = None, cli_args: List[str] = []) -> None:

    # When using new cli, we don't do anything here.
    if is_using_new_cli():
        return

    # TODO: refactor this method into a separate method for each command.
    print(f"Running jetpack with args: {cli_args}")

    # Early return if just one argument is provided (e.g. 'uwsgi').
    if len(cli_args) <= 1:
        return

    """
    Execute a job: `python jetpack_main.py job exec_id qualified_job_symbol [base_64_args]`
    """
    if cli_args[1] == "job" and (len(cli_args) == 4 or len(cli_args) == 5):
        exec_id = cli_args[2]
        target_job_name = cli_args[3]
        params = cli_args[4] if len(cli_args) == 5 else ""
        # Find the OnDemand job, if there is one.
        func = symbols.get_symbol_table()[Symbol(target_job_name)]
        Job(func).exec(exec_id, params)
        return

    """
    DEPRECATED: we now use `job` to run cronjobs.
    This will be removed soon

    Execute a cronjob: `python jetpack_main.py cronjob cronjob_name`
    For legacy reasons, `run` is an alias for `cronjob`. We intend to remove this alias eventually.
    """
    if (cli_args[1] == "run" or cli_args[1] == "cronjob") and len(cli_args) == 3:
        target_job_name = cli_args[2]
        for job in schedule.get_jobs():
            if utils.job_name(job) == target_job_name:
                job.job_func()
                return

        raise JobNotFoundError(f'Target cronjob "{target_job_name}" not found')

    """
    Get all jobs `python jetpack_main.py describe`
    """
    if cli_args[1] == "describe" and len(cli_args) == 2:
        print(json_format.MessageToJson(describe_output(app)))
        return

    """
    Get all jobs `python jetpack_main.py describe-to-redis`
    """
    if cli_args[1] == "describe-to-redis" and len(cli_args) == 4:
        host = cli_args[2]
        key = cli_args[3]
        r = redis.Redis(host=host, port=6379, db=0)
        r.set(key, json_format.MessageToJson(describe_output(app)))
        return


def describe_output(app: Any) -> Any:
    # Should these be the same? The concepts are really similar.
    cron_jobs = []
    jobs = []
    for job in schedule.get_jobs():

        if job.at_time is not None:
            target_time = job.at_time.isoformat()
        else:
            target_time = None

        target_day_of_week: Optional[int] = None
        if job.start_day is not None:
            target_day_of_week = cronjob_pb2.DayOfWeek.Value(job.start_day.upper())

        cron_jobs.append(
            cronjob_pb2.CronJob(
                function=utils.job_name(job),
                target_time=target_time,
                target_day_of_week=target_day_of_week,
                unit=cronjob_pb2.Unit.Value(job.unit.upper()),
                interval=job.interval,
            )
        )

    for symbol in symbols.get_symbol_table().defined_symbols():
        # Note that cronjobs are also jobs and will be added to this list.
        jobs.append(
            job_pb2.Job(
                qualified_symbol=symbol,
            )
        )

    try:
        enum_name = app.__class__.__name__.upper() if app is not None else "NONE"
        framework = describe_pb2.Framework.Value(enum_name)
    except ValueError:
        framework = describe_pb2.Framework.Value("UNKNOWN")

    return describe_pb2.DescribeOutput(
        cron_jobs=cron_jobs,
        jobs=jobs,
        framework=framework,
    )


class JobNotFoundError(Exception):
    pass



import enum
import os
from typing import Any, Awaitable, Dict, Optional, Tuple, Union, cast

import grpc
import jsonpickle

from jetpack import __version__
from jetpack._remote import codec
from jetpack.config import namespace, symbols
from jetpack.proto.runtime.v1alpha1 import remote_pb2, remote_pb2_grpc


@enum.unique
class LaunchJobMode(enum.Enum):
    ASYNC = "ASYNC"
    ASYNC_FIRE_AND_FORGET = "ASYNC_FIRE_AND_FORGET"
    FIRE_AND_FORGET = "FIRE_AND_FORGET"
    BLOCKING = "BLOCKING"


def mode_is_async(mode: LaunchJobMode) -> bool:
    return mode == LaunchJobMode.ASYNC or mode == LaunchJobMode.ASYNC_FIRE_AND_FORGET


def mode_is_fire_and_forget(mode: LaunchJobMode) -> bool:
    return (
        mode == LaunchJobMode.FIRE_AND_FORGET
        or mode == LaunchJobMode.ASYNC_FIRE_AND_FORGET
    )


# Named this is_wait_for_response instead of is_blocking because the latter
# sounds confusing with async. Maybe we should rename BLOCKING to REGULAR?
def mode_is_wait_for_response(mode: LaunchJobMode) -> bool:
    return mode == LaunchJobMode.ASYNC or mode == LaunchJobMode.BLOCKING


class JetpackException(Exception):
    """Base class for exceptions in this module"""

    pass


class RuntimeException(JetpackException):
    """Exception raised for errors in the Jetpack runtime and kubernetes."""

    def __init__(self, message: str) -> None:
        self.message = message


class ApplicationException(JetpackException):
    """Exception raised for errors from application-code that is using the SDK.

    TODO DEV-157
    For exceptions raised by remote functions and jobs, we serialize the
    userland exception in the backend and save it here. The userland exception
    is re-raised by the SDK for the caller of the remote function or job.
    """

    def __init__(self, message: str) -> None:
        self.message = message


class NoControllerAddressError(JetpackException):
    pass


class Client:
    def __init__(self) -> None:
        host = os.environ.get(
            "JETPACK_RUNTIME_SERVICE_HOST",
            "remotesvc.jetpack-runtime.svc.cluster.local",
        )
        port = os.environ.get("JETPACK_RUNTIME_SERVICE_PORT", "80")
        self.address: str = f"{host.strip()}:{port.strip()}"
        self.stub: Optional[remote_pb2_grpc.RemoteExecutorStub] = None
        self.async_stub: Optional[remote_pb2_grpc.RemoteExecutorStub] = None
        self.dry_run = False
        self.dry_run_last_request: Optional[
            Union[
                remote_pb2.LaunchJobRequest,
                remote_pb2.LaunchBlockingJobRequest,
                remote_pb2.PostResultRequest,
            ]
        ] = None

    def dial(self, mode: LaunchJobMode) -> None:
        if not self.address:
            raise NoControllerAddressError("Controller address is not set")
        # Since this is inter-cluster communication, insecure is fine.
        # In the future this won't even leave the pod, and use a sidecar so
        # it will be localhost.

        # TODO(Landau): When/how should we close the channels?
        if mode_is_async(mode):
            # Warning, this calls asyncio.get_event_loop() which will throw
            # exception if event loop has already been set and closed.
            async_channel = grpc.aio.insecure_channel(self.address)
            self.async_stub = remote_pb2_grpc.RemoteExecutorStub(async_channel)
        else:
            channel = grpc.insecure_channel(self.address)
            self.stub = remote_pb2_grpc.RemoteExecutorStub(channel)

    def launch_job(
        self,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Launches a k8s job and blocks until the job completes, before returning.
        For now this function assumes job will live in same namespace where the
        launcher is located.

        Keyword arguments:
        qualified_name -- qualified name as produced by utils.qualified_func_name
        module -- jetpack module where the job resides. Used to determine correct
        docker image.
        """
        response = cast(
            remote_pb2.LaunchBlockingJobResponse,
            self._launch_job(LaunchJobMode.BLOCKING, qualified_name, args, kwargs),
        )

        return self._transform_response_exception(response)

    async def launch_async_job(
        self,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        response = cast(
            Awaitable[remote_pb2.LaunchBlockingJobResponse],
            self._launch_job(LaunchJobMode.ASYNC, qualified_name, args, kwargs),
        )

        return self._transform_response_exception(await response)

    def launch_fire_and_forget_job(
        self,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> remote_pb2.LaunchJobResponse:
        """Launches a k8s job. For now this function assumes job will live in
        same namespace where the launcher is located.

        Keyword arguments:
        qualified_name -- qualified name as produced by utils.qualified_func_name
        module -- jetpack module where the job resides. Used to determine correct
        docker image.
        """
        response = cast(
            remote_pb2.LaunchJobResponse,
            self._launch_job(
                LaunchJobMode.FIRE_AND_FORGET, qualified_name, args, kwargs
            ),
        )
        return response

    async def launch_fire_and_forget_async_job(
        self,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> remote_pb2.LaunchJobResponse:
        response = cast(
            Awaitable[remote_pb2.LaunchJobResponse],
            self._launch_job(
                LaunchJobMode.ASYNC_FIRE_AND_FORGET,
                qualified_name,
                args,
                kwargs,
            ),
        )
        return await response

    def post_result(
        self, exec_id: str, value: Any = None, error: Optional[Exception] = None
    ) -> remote_pb2.PostResultResponse:

        if not exec_id:
            # Note: value=None and error=None is acceptable because a job
            # can run successfully and return nothing.
            raise Exception("An exec_id is required to post a result")

        result = remote_pb2.Result()
        if error:
            result.error.code = remote_pb2.APPLICATION
            result.error.message = str(error)
            result.error.encoded_error = bytes(jsonpickle.encode(error), "utf-8")
        else:  # order matters, as 'value' can be None
            result.value.encoded_value = bytes(jsonpickle.encode(value), "utf-8")

        request = remote_pb2.PostResultRequest(
            exec_id=exec_id,
            result=result,
        )

        if self.dry_run:
            self.dry_run_last_request = request
            print(f"Dry Run:\n{request}")
            return remote_pb2.PostResultResponse()

        self._maybe_dial(LaunchJobMode.BLOCKING)
        assert self.stub is not None

        return cast(remote_pb2.PostResultResponse, self.stub.PostResult(request))

    def _build_request(
        self,
        mode: LaunchJobMode,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]],
        kwargs: Optional[Dict[str, Any]],
    ) -> Union[remote_pb2.LaunchJobRequest, remote_pb2.LaunchBlockingJobRequest]:
        encoded_args = b""
        if args or kwargs:
            encoded_args = codec.encode_args(
                args if args else None,
                kwargs if kwargs else None,
            ).encode("utf-8")

        current_namespace = namespace.get()
        job = remote_pb2.RemoteJob(
            qualified_symbol=qualified_name,
            encoded_args=encoded_args,
            hostname=os.environ["HOSTNAME"],  # k8s sets this
        )
        if current_namespace:
            job.namespace = current_namespace

        request: Union[remote_pb2.LaunchJobRequest, remote_pb2.LaunchBlockingJobRequest]

        if mode_is_fire_and_forget(mode):
            request = remote_pb2.LaunchJobRequest(job=job, sdk_version=__version__)
        elif mode_is_wait_for_response(mode):
            request = remote_pb2.LaunchBlockingJobRequest(
                job=job, sdk_version=__version__
            )

        else:
            raise Exception(f"unsupported mode: {mode}")

        if self.dry_run:
            self.dry_run_last_request = request

        return request

    def _launch_job(
        self,
        mode: LaunchJobMode,
        qualified_name: str,
        args: Optional[Tuple[Any, ...]],
        kwargs: Optional[Dict[str, Any]],
    ) -> Union[
        # TODO: Length of this return type makes me thing we should separate
        # the different cases into different implementations.
        remote_pb2.LaunchJobResponse,
        Awaitable[remote_pb2.LaunchJobResponse],
        remote_pb2.LaunchBlockingJobResponse,
        Awaitable[remote_pb2.LaunchBlockingJobResponse],
        None,  # Needed for dry run
        Awaitable[None],  # Needed for dry run
    ]:
        request = self._build_request(mode, qualified_name, args, kwargs)

        # TODO(Landau): This is gross, remove dry_run.
        if self.dry_run:
            if mode_is_async(mode):

                async def empty_awaitable() -> None:
                    pass

                return empty_awaitable()
            return None

        self._maybe_dial(mode)

        if mode == LaunchJobMode.FIRE_AND_FORGET:
            assert self.stub is not None
            return cast(remote_pb2.LaunchJobResponse, self.stub.LaunchJob(request))
        elif mode == LaunchJobMode.BLOCKING:
            assert self.stub is not None
            return cast(
                remote_pb2.LaunchBlockingJobResponse,
                self.stub.LaunchBlockingJob(request),
            )

        # Async flavors return a coroutine.
        elif mode == LaunchJobMode.ASYNC:
            assert self.async_stub is not None
            return cast(
                Awaitable[remote_pb2.LaunchBlockingJobResponse],
                self.async_stub.LaunchBlockingJob(request),
            )
        elif mode == LaunchJobMode.ASYNC_FIRE_AND_FORGET:
            assert self.async_stub is not None
            return cast(
                Awaitable[remote_pb2.LaunchJobResponse],
                self.async_stub.LaunchJob(request),
            )
        else:
            raise Exception(f"unsupported mode: {mode}")

    @staticmethod
    def _transform_response_exception(
        response: remote_pb2.LaunchBlockingJobResponse,
    ) -> Any:
        if not isinstance(response, remote_pb2.LaunchBlockingJobResponse):
            # can happen for dry-run
            # TODO this is a code-smell. We should instead properly mock
            # the response from the grpc call.
            return

        if response.result.HasField("error") and response.result.error.encoded_error:
            e = jsonpickle.decode(response.result.error.encoded_error)
            raise e
        elif response.result.HasField("value"):
            val = jsonpickle.decode(response.result.value.encoded_value)
            return val
        else:
            raise RuntimeException(
                f"Either 'value' or 'error' should be specified in response. Got: {response}"
            )

    def _maybe_dial(self, mode: LaunchJobMode) -> None:
        if not mode_is_async(mode) and self.stub is None:
            self.dial(mode)
        elif mode_is_async(mode) and self.async_stub is None:
            self.dial(mode)



import importlib.abc
import importlib.util
import os
import sys
from typing import cast


def load_user_entrypoint(entrypoint: str) -> None:
    # We could use importlib.import_module but we want to guarantee entrypoint
    # is loaded with the correct module name. (e.g. main.py should be module main)
    # See jetpack.utils.qualified_name for how we generage job names
    # In particular, app/main.py should be module "main" and not "app.main"

    # We ensure the entrypoint path is in path so that imports work as expected.
    sys.path.append(os.path.dirname(entrypoint))
    module_name = os.path.basename(entrypoint).split(".")[0]
    spec = importlib.util.spec_from_file_location(module_name, entrypoint)
    assert spec is not None
    entrypoint_module = importlib.util.module_from_spec(spec)
    cast(importlib.abc.Loader, spec.loader).exec_module(entrypoint_module)



from typing import Any, Callable

from jetpack._job.job import Job as _Job
from jetpack.config import symbols


class JobDecorator:
    def __call__(self, fn: Callable[..., Any]) -> Callable[..., Any]:
        job = _Job(fn)
        symbols.get_symbol_table().register(fn)
        return job


# @job will most likely be deprecated in favor of @function.
job = JobDecorator()

# @function is our general remote work decorator. It does not specify how the
# work will be done (RPC, job, queue, etc) and instead leaves that as an
# implementation detail.
# For now, we alias @job, but once/if we remove @job we can refactor other
# internal code
function = JobDecorator()



from __future__ import annotations

import asyncio
import base64
import inspect
import os
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, Union, ValuesView

from jetpack import utils
from jetpack._job.client import Client
from jetpack._remote import codec
from jetpack.proto.runtime.v1alpha1 import remote_pb2

client = Client()


def get_client() -> Client:
    return client


class AlreadyExistsError(Exception):
    pass


class Job:
    # Question: do we need a separate Dict for blocking jobs?
    jobs: Dict[str, Job] = {}

    def __init__(self, func: Callable[..., Any]) -> None:
        self.func = func

    def __call__(self, *args: Any, **kwargs: Any) -> Union[Any, Awaitable[Any]]:
        """
        No return value for now, but eventually this will return a handle
        or an awaitable
        """
        if inspect.iscoroutinefunction(self.func):
            return client.launch_async_job(self.name(), args, kwargs)
        else:
            return client.launch_job(self.name(), args, kwargs)

    def fire_and_forget(
        self, *args: Any, **kwargs: Any
    ) -> Union[remote_pb2.LaunchJobResponse, Awaitable[remote_pb2.LaunchJobResponse]]:
        if inspect.iscoroutinefunction(self.func):
            return client.launch_fire_and_forget_async_job(self.name(), args, kwargs)
        else:
            return client.launch_fire_and_forget_job(self.name(), args, kwargs)

    def name(self) -> str:
        return utils.qualified_func_name(self.func)

    def exec(self, exec_id: str, base64_encoded_args: str = "") -> None:

        args: Tuple[Any, ...] = ()
        kwargs: Dict[str, Any] = {}
        if base64_encoded_args:
            encoded_args = base64.b64decode(base64_encoded_args).decode("utf-8")
            decoded_args, decoded_kwargs = codec.decode_args(encoded_args)
            if decoded_args:
                args = decoded_args
            if decoded_kwargs:
                kwargs = decoded_kwargs

        retval, err = None, None
        try:
            if inspect.iscoroutinefunction(self.func):
                retval = asyncio.run(self.func(*args, **kwargs))
            else:
                retval = self.func(*args, **kwargs)
        except Exception as e:
            err = e

        # for now, we post the result back to the remotesvc. A slightly better approach is to
        # have the caller of this function post it (the CLI). Doing it here for now because
        # the remotesvc is already initialized and set up here.
        client.post_result(exec_id, value=retval, error=err)
